input[type='color'],
input[type='date'],
input[type='datetime-local'],
input[type='datetime'],
input[type='email'],
input[type='month'],
input[type='number'],
input[type='password'],
input[type='search'],
input[type='tel'],
input[type='text'],
input[type='time'],
input[type='url'],
input[type='week'] {
  box-shadow: none;
}

.theme-switcher {
  border-radius: 50%;
  position: fixed;
  right: 1.6em;
  bottom: 1.4em;
  z-index: 3;
  border: none;
  height: 2.2em;
  width: 2.2em;
  background-color: #fcfcfc;
  font-size: 20px;
  -webkit-box-shadow: 0px 3px 14px 4px rgba(0, 0, 0, 0.62);
  box-shadow: 0px 3px 14px 4px rgba(0, 0, 0, 0.62);
  color: #404040;
  transition: all 0.3s ease-in-out;
}

body,
.wy-nav-content-wrap,
.wy-nav-content,
.section,
.highlight,
.rst-content div[class^='highlight'],
.wy-nav-content a,
.btn-neutral,
.btn,
footer,
.wy-nav-side,
.wy-menu-vertical li,
.wy-menu-vertical a,
.wy-side-nav-search .wy-dropdown,
.wy-side-nav-search a,
.wy-side-nav-search input,
html.writer-html4 .rst-content dl:not(.docutils) > dt,
html.writer-html5
  .rst-content
  dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.glossary):not(.simple)
  > dt,
.rst-content code,
.rst-content tt,
html.writer-html4 .rst-content dl:not(.docutils) dl:not(.field-list) > dt,
html.writer-html5
  .rst-content
  dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.glossary):not(.simple)
  dl:not(.field-list)
  > dt,
code,
.rst-content code.xref,
.rst-content tt.xref {
}
